# Copyright (c) Facebook, Inc. and its affiliates.
import colorsys
import logging
import math
from enum import Enum, unique

import cv2
import matplotlib as mpl
import matplotlib.colors as mplc
import matplotlib.figure as mplfigure
import numpy as np
import pycocotools.mask as mask_util
import torch
from detectron2.data import MetadataCatalog
from detectron2.structures import (BitMasks, Boxes, BoxMode, Keypoints,
                                   PolygonMasks, RotatedBoxes)
from detectron2.utils.file_io import PathManager
from detectron2.utils.visualizer import ColorMode, GenericMask, Visualizer
from matplotlib.backends.backend_agg import FigureCanvasAgg
from PIL import Image

logger = logging.getLogger(__name__)


def _create_text_labels(classes, scores, class_names, is_crowd=None, args=None):
    """
    Args:
        classes (list[int] or None):
        scores (list[float] or None):
        class_names (list[str] or None):
        is_crowd (list[bool] or None):

    Returns:
        list[str] or None
    """
    replace_set = [
        ["monitor_(computer_equipment) computer_monitor", "computer_monitor"],
        ["_(computer_equipment)", ""],
        ["_(automobile)", ""],
        ["_(drink container)", ""],
    ]
    labels = None
    if classes is not None:
        if args.MODEL.EVAL_PROPOSAL:
            labels = ["" for _ in classes]
        elif class_names is not None and len(class_names) > 0:
            labels = []
            for i in classes:
                class_name = class_names[i]
                for long_name, short_name in replace_set:
                    class_name = class_name.replace(long_name, short_name)
                labels.append(class_name)
            # labels = [class_names[i] for i in classes]
        else:
            labels = [str(i) for i in classes]
    if scores is not None:
        if labels is None:
            labels = ["{:.0f}%".format(s * 100) for s in scores]
        else:
            labels = ["{} {:.0f}%".format(l, s * 100) for l, s in zip(labels, scores)]
    if labels is not None and is_crowd is not None:
        labels = [l + ("|crowd" if crowd else "") for l, crowd in zip(labels, is_crowd)]
    return labels


class CustomVisualizer(Visualizer):
    """
    Support to visualize the output of different supervision types
    """

    def draw_instance_predictions(self, predictions, args):
        """
        Draw instance-level prediction results on an image.

        Args:
            predictions (Instances): the output of an instance detection/segmentation
                model. Following fields will be used to draw:
                "pred_boxes", "pred_classes", "scores", "pred_masks" (or "pred_masks_rle").

        Returns:
            output (VisImage): image object with visualizations.
        """
        boxes = predictions.pred_boxes if predictions.has("pred_boxes") else None
        scores = predictions.scores if predictions.has("scores") else None
        scores = None if not args.VIS.SCORE else scores
        boxes = None if not args.VIS.BOX else boxes

        classes = (
            predictions.pred_classes.tolist()
            if predictions.has("pred_classes")
            else None
        )
        labels = _create_text_labels(
            classes, scores, self.metadata.get("thing_classes", None), args=args
        )
        labels = None if not args.VIS.LABELS else labels
        keypoints = (
            predictions.pred_keypoints if predictions.has("pred_keypoints") else None
        )

        if predictions.has("pred_masks"):
            masks = np.asarray(predictions.pred_masks)
            masks = [
                GenericMask(x, self.output.height, self.output.width) for x in masks
            ]
        else:
            masks = None
        masks = None if not args.VIS.MASK else masks

        if self._instance_mode == ColorMode.SEGMENTATION and self.metadata.get(
            "thing_colors"
        ):
            colors = [
                self._jitter([x / 255 for x in self.metadata.thing_colors[c]])
                for c in classes
            ]
            alpha = 0.8
        else:
            colors = None
            alpha = 0.5

        if self._instance_mode == ColorMode.IMAGE_BW:
            self.output.reset_image(
                self._create_grayscale_image(
                    (predictions.pred_masks.any(dim=0) > 0).numpy()
                    if predictions.has("pred_masks")
                    else None
                )
            )
            alpha = 0.3

        self.overlay_instances(
            masks=masks,
            boxes=boxes,
            labels=labels,
            keypoints=keypoints,
            assigned_colors=colors,
            alpha=alpha,
        )
        return self.output

    def draw_box(self, box_coord, alpha=0.9, edge_color="g", line_style="-"):
        """ """
        x0, y0, x1, y1 = box_coord
        width = x1 - x0
        height = y1 - y0

        linewidth = max(self._default_font_size / 4, 1)

        self.output.ax.add_patch(
            mpl.patches.Rectangle(
                (x0, y0),
                width,
                height,
                fill=False,
                edgecolor=edge_color,
                linewidth=linewidth * self.output.scale,
                alpha=alpha,
                linestyle=line_style,
            )
        )
        return self.output
